@@ -1,5 +1,5 @@
 /* abootimg -  Manipulate (read, modify, create) Android Boot Images
- * Copyright (c) 2010 Gilles Grandou <gilles@grandou.net>
+ * Copyright (c) 2010-2011 Gilles Grandou <gilles@grandou.net>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -24,13 +24,34 @@
 #include <stdarg.h>
 #include <fcntl.h>
 #include <sys/stat.h>
+#include <unistd.h>
+
+
+#ifdef __linux__
 #include <sys/ioctl.h>
-#include <linux/fs.h>
+#include <linux/fs.h> /* BLKGETSIZE64 */
+#endif
+
+#ifdef __CYGWIN__
+#include <sys/ioctl.h>
+#include <cygwin/fs.h> /* BLKGETSIZE64 */
+#endif
+
+#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
+#include <sys/disk.h> /* DIOCGMEDIASIZE */
+#include <sys/sysctl.h>
+#endif
+
+#if defined(__APPLE__)
+# include <sys/disk.h> /* DKIOCGETBLOCKCOUNT */
+#endif
+
 
 #ifdef HAS_BLKID
 #include <blkid/blkid.h>
 #endif
 
+#include "version.h"
 #include "bootimg.h"
 
 
@@ -87,10 +108,33 @@ void abort_printf(char *fmt, ...)
 }
 
 
+int blkgetsize(int fd, unsigned long long *pbsize)
+{
+# if defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
+  return ioctl(fd, DIOCGMEDIASIZE, pbsize);
+# elif defined(__APPLE__)
+  return ioctl(fd, DKIOCGETBLOCKCOUNT, pbsize);
+# elif defined(__NetBSD__)
+  // does a suitable ioctl exist?
+  // return (ioctl(fd, DIOCGDINFO, &label) == -1);
+  return 1;
+# elif defined(__linux__) || defined(__CYGWIN__)
+  return ioctl(fd, BLKGETSIZE64, pbsize);
+# elif defined(__GNU__)
+  // does a suitable ioctl for HURD exist?
+  return 1;
+# else
+  return 1;
+# endif
+
+}
+
 void print_usage(void)
 {
   printf (
  " abootimg - manipulate Android Boot Images.\n"
+ " (c) 2010-2011 Gilles Grandou <gilles@grandou.net>\n"
+ " " VERSION_STR "\n"
  "\n"
  " abootimg [-h]\n"
  "\n"
@@ -231,24 +275,37 @@ enum command parse_args(int argc, char** argv, t_abootimg* img)
 
 int check_boot_img_header(t_abootimg* img)
 {
-  if (strncmp(img->header.magic, BOOT_MAGIC, BOOT_MAGIC_SIZE))
+  if (strncmp(img->header.magic, BOOT_MAGIC, BOOT_MAGIC_SIZE)) {
+    fprintf(stderr, "%s: no Android Magic Value\n", img->fname);
     return 1;
+  }
 
-  if (!(img->header.kernel_size))
+  if (!(img->header.kernel_size)) {
+    fprintf(stderr, "%s: kernel size is null\n", img->fname);
     return 1;
+  }
 
-  if (!(img->header.ramdisk_size))
+  if (!(img->header.ramdisk_size)) {
+    fprintf(stderr, "%s: ramdisk size is null\n", img->fname);
     return 1;
+  }
 
   unsigned page_size = img->header.page_size;
+  if (!page_size) {
+    fprintf(stderr, "%s: Image page size is null\n", img->fname);
+    return 1;
+  }
+
   unsigned n = (img->header.kernel_size + page_size - 1) / page_size;
   unsigned m = (img->header.ramdisk_size + page_size - 1) / page_size;
   unsigned o = (img->header.second_size + page_size - 1) / page_size;
 
   unsigned total_size = (1+n+m+o)*page_size;
 
-  if (total_size > img->size)
+  if (total_size > img->size) {
+    fprintf(stderr, "%s: sizes mismatches in boot image\n", img->fname);
     return 1;
+  }
 
   return 0;
 }
@@ -278,7 +335,7 @@ void check_if_block_device(t_abootimg* img)
       abort_perror(img->fname);
     
     unsigned long long bsize = 0;
-    if (ioctl(fd, BLKGETSIZE64, &bsize))
+    if (blkgetsize(fd, &bsize))
       abort_perror(img->fname);
     img->size = bsize;
 
@@ -300,9 +357,8 @@ void open_bootimg(t_abootimg* img, char* mode)
 
 void read_header(t_abootimg* img)
 {
-
-  fread(&img->header, sizeof(boot_img_hdr), 1, img->stream);
-  if (ferror(img->stream))
+  size_t rb = fread(&img->header, sizeof(boot_img_hdr), 1, img->stream);
+  if ((rb!=1) || ferror(img->stream))
     abort_perror(img->fname);
   else if (feof(img->stream))
     abort_printf("%s: cannot read image header\n", img->fname);
@@ -314,7 +370,8 @@ void read_header(t_abootimg* img)
 
   if (S_ISBLK(s.st_mode)) {
     unsigned long long bsize = 0;
-    if (ioctl(fd, BLKGETSIZE64, &bsize))
+
+    if (blkgetsize(fd, &bsize))
       abort_perror(img->fname);
     img->size = bsize;
     img->is_blkdev = 1;
@@ -409,7 +466,7 @@ void update_header(t_abootimg* img)
     printf("reading config file %s\n", img->config_fname);
 
     char* line = NULL;
-    unsigned len = 0;
+    size_t len = 0;
     int read;
 
     while ((read = getline(&line, &len, config_file)) != -1) {
@@ -430,7 +487,7 @@ void update_header(t_abootimg* img)
     printf("reading config args\n");
 
     char* line = NULL;
-    unsigned len = 0;
+    size_t len = 0;
     int read;
 
     while ((read = getline(&line, &len, config_file)) != -1) {
@@ -452,6 +509,9 @@ void update_images(t_abootimg *img)
   unsigned rsize = img->header.ramdisk_size;
   unsigned ssize = img->header.second_size;
 
+  if (!page_size)
+    abort_printf("%s: Image page size is null\n", img->fname);
+
   unsigned n = (ksize + page_size - 1) / page_size;
   unsigned m = (rsize + page_size - 1) / page_size;
   unsigned o = (ssize + page_size - 1) / page_size;
@@ -471,8 +531,8 @@ void update_images(t_abootimg *img)
     char* k = malloc(ksize);
     if (!k)
       abort_perror("");
-    fread(k, 1, ksize, stream);
-    if (ferror(stream))
+    size_t rb = fread(k, ksize, 1, stream);
+    if ((rb!=1) || ferror(stream))
       abort_perror(img->kernel_fname);
     else if (feof(stream))
       abort_printf("%s: cannot read kernel\n", img->kernel_fname);
@@ -493,8 +553,8 @@ void update_images(t_abootimg *img)
     char* r = malloc(rsize);
     if (!r)
       abort_perror("");
-    fread(r, 1, rsize, stream);
-    if (ferror(stream))
+    size_t rb = fread(r, rsize, 1, stream);
+    if ((rb!=1) || ferror(stream))
       abort_perror(img->ramdisk_fname);
     else if (feof(stream))
       abort_printf("%s: cannot read ramdisk\n", img->ramdisk_fname);
@@ -509,8 +569,8 @@ void update_images(t_abootimg *img)
       abort_perror("");
     if (fseek(img->stream, roffset, SEEK_SET))
       abort_perror(img->fname);
-    fread(r, 1, rsize, img->stream);
-    if (ferror(img->stream))
+    size_t rb = fread(r, rsize, 1, img->stream);
+    if ((rb!=1) || ferror(img->stream))
       abort_perror(img->fname);
     else if (feof(img->stream))
       abort_printf("%s: cannot read ramdisk\n", img->fname);
@@ -529,8 +589,8 @@ void update_images(t_abootimg *img)
     char* s = malloc(ssize);
     if (!s)
       abort_perror("");
-    fread(s, 1, ssize, stream);
-    if (ferror(stream))
+    size_t rb = fread(s, ssize, 1, stream);
+    if ((rb!=1) || ferror(stream))
       abort_perror(img->second_fname);
     else if (feof(stream))
       abort_printf("%s: cannot read second stage\n", img->second_fname);
@@ -545,8 +605,8 @@ void update_images(t_abootimg *img)
       abort_perror("");
     if (fseek(img->stream, soffset, SEEK_SET))
       abort_perror(img->fname);
-    fread(s, 1, ssize, img->stream);
-    if (ferror(img->stream))
+    size_t rb = fread(s, ssize, 1, img->stream);
+    if ((rb!=1) || ferror(img->stream))
       abort_perror(img->fname);
     else if (feof(img->stream))
       abort_printf("%s: cannot read second stage\n", img->fname);
@@ -591,16 +651,16 @@ void write_bootimg(t_abootimg* img)
   if (ferror(img->stream))
     abort_perror(img->fname);
 
-  fwrite(padding, 1, psize - sizeof(img->header), img->stream);
+  fwrite(padding, psize - sizeof(img->header), 1, img->stream);
   if (ferror(img->stream))
     abort_perror(img->fname);
 
   if (img->kernel) {
-    fwrite(img->kernel, 1, img->header.kernel_size, img->stream);
+    fwrite(img->kernel, img->header.kernel_size, 1, img->stream);
     if (ferror(img->stream))
       abort_perror(img->fname);
 
-    fwrite(padding, 1, psize - (img->header.kernel_size % psize), img->stream);
+    fwrite(padding, psize - (img->header.kernel_size % psize), 1, img->stream);
     if (ferror(img->stream))
       abort_perror(img->fname);
   }
@@ -609,11 +669,11 @@ void write_bootimg(t_abootimg* img)
     if (fseek(img->stream, (1+n)*psize, SEEK_SET))
       abort_perror(img->fname);
 
-    fwrite(img->ramdisk, 1, img->header.ramdisk_size, img->stream);
+    fwrite(img->ramdisk, img->header.ramdisk_size, 1, img->stream);
     if (ferror(img->stream))
       abort_perror(img->fname);
 
-    fwrite(padding, 1, psize - (img->header.ramdisk_size % psize), img->stream);
+    fwrite(padding, psize - (img->header.ramdisk_size % psize), 1, img->stream);
     if (ferror(img->stream))
       abort_perror(img->fname);
   }
@@ -622,11 +682,11 @@ void write_bootimg(t_abootimg* img)
     if (fseek(img->stream, (1+n+m)*psize, SEEK_SET))
       abort_perror(img->fname);
 
-    fwrite(img->second, 1, img->header.second_size, img->stream);
+    fwrite(img->second, img->header.second_size, 1, img->stream);
     if (ferror(img->stream))
       abort_perror(img->fname);
 
-    fwrite(padding, 1, psize - (img->header.second_size % psize), img->stream);
+    fwrite(padding, psize - (img->header.second_size % psize), 1, img->stream);
     if (ferror(img->stream))
       abort_perror(img->fname);
   }
@@ -717,8 +777,8 @@ void extract_kernel(t_abootimg* img)
   if (fseek(img->stream, psize, SEEK_SET))
     abort_perror(img->fname);
 
-  fread(k, ksize, 1, img->stream);
-  if (ferror(img->stream))
+  size_t rb = fread(k, ksize, 1, img->stream);
+  if ((rb!=1) || ferror(img->stream))
     abort_perror(img->fname);
  
   FILE* kernel_file = fopen(img->kernel_fname, "w");
@@ -746,15 +806,15 @@ void extract_ramdisk(t_abootimg* img)
 
   printf ("extracting ramdisk in %s\n", img->ramdisk_fname);
 
-  void* r = malloc(ksize);
+  void* r = malloc(rsize);
   if (!r) 
     abort_perror(NULL);
 
   if (fseek(img->stream, roffset, SEEK_SET))
     abort_perror(img->fname);
 
-  fread(r, rsize, 1, img->stream);
-  if (ferror(img->stream))
+  size_t rb = fread(r, rsize, 1, img->stream);
+  if ((rb!=1) || ferror(img->stream))
     abort_perror(img->fname);
  
   FILE* ramdisk_file = fopen(img->ramdisk_fname, "w");
@@ -793,8 +853,8 @@ void extract_second(t_abootimg* img)
   if (fseek(img->stream, soffset, SEEK_SET))
     abort_perror(img->fname);
 
-  fread(s, ssize, 1, img->stream);
-  if (ferror(img->stream))
+  size_t rb = fread(s, ssize, 1, img->stream);
+  if ((rb!=1) || ferror(img->stream))
     abort_perror(img->fname);
  
   FILE* second_file = fopen(img->second_fname, "w");
@@ -825,6 +885,7 @@ t_abootimg* new_bootimg()
   img->second_fname = "stage2.img";
 
   memcpy(img->header.magic, BOOT_MAGIC, BOOT_MAGIC_SIZE);
+  img->header.page_size = 2048;  // a sensible default page size
 
   return img;
 }
